package cmd

import (
	"bytes"
	"regexp"

	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	yaml "gopkg.in/yaml.v2"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
)

func addPipelineFilterFlags(flags *pflag.FlagSet, defaultIncludeDraft bool) {
	flags.Bool("status-inactive", false, "Include inactive versions")
	flags.Bool("status-draft", defaultIncludeDraft, "Include draft versions")
	flags.Bool("all", false, "Include all versions")
	flags.String("name", "", "Include pipeline with name matching the given regex")
	flags.String("version", "", "Include pipeline with version matching the given version")
	flags.String(
		"actor", "",
		"Include pipeline with actor matching the given actor name, including by roles")
	flags.String(
		"actor-role", "",
		"Include pipeline with a role matching the given actor role name")
	flags.String(
		"account", "",
		"Include pipeline with any of the account actor, including by roles")
	flags.String(
		"msgtype", "",
		"Include pipeline with a matching msgtype list including the given type")
}

type pipelineFilterFlags struct {
	StatusDraft    bool
	StatusInactive bool
	Name           *regexp.Regexp
	Version        string
	Actor          string
	ActorRole      string
	Account        string
	MsgType        string
}

func parsePipelineFilterFlags(cmd *cobra.Command) (*pipelineFilterFlags, error) {
	var (
		nameRe *regexp.Regexp
		all    = cmd.Flag("all").Value.String() == "true"
	)
	if name := cmd.Flag("name").Value.String(); name != "" {
		r, err := regexp.Compile(name)
		if err != nil {
			return nil, err
		}
		nameRe = r
	}
	return &pipelineFilterFlags{
		StatusDraft:    all || cmd.Flag("status-draft").Value.String() == "true",
		StatusInactive: all || cmd.Flag("status-inactive").Value.String() == "true",
		Name:           nameRe,
		Version:        cmd.Flag("version").Value.String(),
		Actor:          cmd.Flag("actor").Value.String(),
		ActorRole:      cmd.Flag("actor-role").Value.String(),
		Account:        cmd.Flag("account").Value.String(),
		MsgType:        cmd.Flag("msgtype").Value.String(),
	}, nil
}

func pipelineFilterMatchActor(
	actors []api.Actor, graph storage.Graph, actor api.Actor) bool {
	for _, node := range graph.Nodes {
		for _, name := range node.Actors {
			if actor.Name == name || actor.ID.String() == name {
				return true
			}
		}
		if len(node.Roles) == 1 && len(actor.Roles) == 1 && node.Roles[0] == actor.Roles[0] {
			return true
		}
		if (node.Type == storage.EmitterNode || node.Type == storage.RequestorNode) &&
			actor.Kind == api.Actor_EMITTER &&
			len(node.ActorIDs) == 0 &&
			len(node.Actors) == 0 &&
			len(node.Roles) == 0 {
			return true
		}
	}
	return false
}

func pipelineFilterMatch(
	c *xbus.Client,
	accounts []api.Account, actors []api.Actor,
	filter *pipelineFilterFlags, pipeline api.PipelineInfo,
) (bool, *storage.Graph, error) {
	var (
		graph *storage.Graph
	)
	if filter == nil {
		return true, nil, nil
	}
	if !filter.StatusInactive && pipeline.Status == api.PipelineInfo_INACTIVE {
		return false, nil, nil
	}
	if !filter.StatusDraft && pipeline.Status == api.PipelineInfo_DRAFT {
		return false, nil, nil
	}
	if filter.Name != nil {
		if !filter.Name.MatchString(pipeline.Name) {
			return false, nil, nil
		}
	}
	if filter.Version != "" {
		if pipeline.Version != filter.Version {
			return false, nil, nil
		}
	}
	if filter.Actor != "" ||
		filter.ActorRole != "" ||
		filter.Account != "" ||
		filter.MsgType != "" {
		b := bytes.NewBuffer(nil)
		if err := pipelineGet(c, pipeline.Name, pipeline.Version, b); err != nil {
			return false, nil, err
		}
		graph = &storage.Graph{}
		if err := yaml.Unmarshal(b.Bytes(), graph); err != nil {
			return false, nil, err
		}
	}

	if filter.ActorRole != "" {
		var match bool
		for _, node := range graph.Nodes {
			if len(node.Roles) == 1 && node.Roles[0] == filter.ActorRole {
				match = true
				break
			}
		}
		if !match {
			return false, graph, nil
		}
	}

	if filter.Actor != "" {
		var actor api.Actor
		for _, a := range actors {
			if a.Name == filter.Actor || a.ID.String() == filter.Actor {
				actor = a
			}
		}
		if !pipelineFilterMatchActor(actors, *graph, actor) {
			return false, graph, nil
		}
	}

	if filter.Account != "" {
		var account api.Account
		for _, a := range accounts {
			if a.Name == filter.Account || a.ID.String() == filter.Account {
				account = a
			}
		}
		var match bool
		for _, actor := range actors {
			if actor.AccountID == account.ID {
				if pipelineFilterMatchActor(actors, *graph, actor) {
					match = true
					break
				}
			}
		}
		if !match {
			return false, graph, nil
		}
	}

	if filter.MsgType != "" {
		var match bool

		for _, node := range graph.Nodes {
			if node.Type == storage.EmitterNode || node.Type == storage.RequestorNode {
				if node.SourceMatch != nil {
					for _, t := range node.SourceMatch.EventTypes {
						if t == filter.MsgType {
							match = true
							break
						}
					}
				}
			}
		}

		if !match {
			return false, graph, nil
		}
	}
	return true, graph, nil
}

func filterPipelineList(c *xbus.Client, filter *pipelineFilterFlags) ([]api.PipelineInfo, []*storage.Graph, error) {
	fullList, err := c.PipelineQuery("")
	if err != nil {
		return nil, nil, err
	}
	if filter == nil {
		return fullList, nil, nil
	}

	var (
		accounts []api.Account
		actors   []api.Actor
	)

	if filter.Account != "" {
		list, err := c.AccountList()
		if err != nil {
			return nil, nil, err
		}
		accounts = list
	}
	if filter.Account != "" || filter.Actor != "" {
		list, err := c.ActorList(nil)
		if err != nil {
			return nil, nil, err
		}
		actors = list
	}
	var (
		pipelines []api.PipelineInfo
		graphs    []*storage.Graph
	)
	for _, p := range fullList {
		match, graph, err := pipelineFilterMatch(c, accounts, actors, filter, p)
		if err != nil {
			return nil, nil, err
		}
		if match {
			pipelines = append(pipelines, p)
			graphs = append(graphs, graph)
		}
	}
	return pipelines, graphs, nil
}
