package fullenv

import (
	"io/ioutil"
	"os"
	"regexp"
	"strings"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/testutils"
)

// TestPipelineGet tests the 'xbusctl pipeline get' command
func TestPipelineGet(t *testing.T) {
	env, _ := NewFullEnvForTest(t, "pipelinereload", "pipelinereload.yaml")

	if err := env.Init(); err != nil {
		t.Fatal(err)
	}
	defer env.Close()

	if err := env.Startup(); err != nil {
		t.Fatal(err)
	}

	f1, err := ioutil.TempFile("", "xbus-test")
	if err != nil {
		t.Fatal(err)
	}
	f1.Close()
	defer os.Remove(f1.Name())

	f2, err := ioutil.TempFile("", "xbus-test")
	if err != nil {
		t.Fatal(err)
	}
	f2.Close()
	defer os.Remove(f2.Name())

	env.XbusCtl.MustRun("pipeline", "get", "demo", "--output", f1.Name())
	env.XbusCtl.MustRun("pipeline", "get", "demo", "good", "--output", f2.Name())

	// The two pipeline should be different
	g1, err := ioutil.ReadFile(f1.Name())
	require.NoError(t, err)
	g2, err := ioutil.ReadFile(f2.Name())
	require.NoError(t, err)

	assert.NotEqual(t, string(g1), string(g2))
}

// TestPipelineReload makes sure a freshly activated pipeline matches incoming
// envelopes as it should
func TestPipelineReload(t *testing.T) {
	env, log := NewFullEnvForTest(t, "pipelinereload", "pipelinereload.yaml")

	if err := env.Init(); err != nil {
		t.Fatal(err)
	}

	defer env.Close()
	log.ExpectsError(
		testutils.ExpectsContains("Error sending envelope: No matching pipeline"))

	if err := env.Startup(); err != nil {
		t.Fatal(err)
	}

	sender := env.Emitter.Actor.OpenOutput(
		nil, "default", "demo.count")
	must(sender.AddItems("demo.count", api.Item("item")))
	err := sender.Close()
	if err == nil {
		t.Fatal("emit should fail with 'No matching pipeline'")
	}
	if !strings.HasSuffix(err.Error(), "No matching pipeline") {
		t.Fatal(err)
	}

	// EXTRACT PIDS
	re := regexp.MustCompile(`\| ([0-9a-f-]+) \| [^\|]+ \| [^\|]+ \| [^\|]+ \| [A-Z]+ +\|`)

	matchLines, err := env.XbusCtl.RunReadAndFindSubmatch(
		re, "ps", "--status-running")
	must(err)
	assert.Equal(t, 0, len(matchLines))

	matchLines, err = env.XbusCtl.RunReadAndFindSubmatch(
		re, "ps", "--status-error")
	must(err)
	assert.Equal(t, 1, len(matchLines))

	// now activate the right version of the pipeline
	env.XbusCtl.MustRun("pipeline", "status", "demo", "good", "--active")

	sender = env.Emitter.Actor.OpenOutput(
		nil, "default", "demo.count")
	must(sender.AddItems("demo.count", api.Item("item")))
	must(sender.Close())

	// There is a very short time where the process is running
	time.Sleep(100 * time.Millisecond)

	matchLines, err = env.XbusCtl.RunReadAndFindSubmatch(
		re, "ps", "--status-running")
	must(err)
	assert.Equal(t, 0, len(matchLines))

	matchLines, err = env.XbusCtl.RunReadAndFindSubmatch(
		re, "ps", "--status-error")
	must(err)
	assert.Equal(t, 1, len(matchLines))

	matchLines, err = env.XbusCtl.RunReadAndFindSubmatch(
		re, "ps", "--status-done")
	must(err)
	assert.Equal(t, 1, len(matchLines))
}
