package fullstack

import (
	"context"
	"fmt"
	"testing"
	"time"

	nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"gopkg.in/yaml.v2"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"

	"xbus.io/xbus/testutils"
)

type ProcessExportData struct {
	Logs []api.LogEntry
}

func TestServerTooBusy(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	var hitCount int

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	csub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			hitCount++
			if hitCount > 2 {
				return nil
			}
			return &nrpc.Error{
				Type:    nrpc.Error_SERVERTOOBUSY,
				Message: fmt.Sprintf("hitCount %d", hitCount),
			}
		})
	if err != nil {
		t.Fatal(err)
	}
	defer testutils.DeferedAssertNoError(t, csub.Unsubscribe)

	ft.InstallGraph(`
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> target.default
`)
	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer testutils.DeferedAssertNoError(t, sub.Unsubscribe)

	status, err := emitter.SendItems(
		nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	timeout, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()
	var (
		pStatus api.Process_Status
		pID     api.UUID
	)
	for pStatus != api.Process_DONE && pStatus != api.Process_ERROR {
		select {
		case <-timeout.Done():
			t.Fatalf("Process never ended")
		case eState := <-pStateChan:
			t.Logf("New envelope status: %s", eState.Status)
			pID = eState.ProcessID
			pStatus = eState.Status
		}
	}

	require.Equal(t, api.Process_DONE.String(), pStatus.String())
	// Export the process, check the logs
	var b []byte
	require.NoError(t, ft.Client.Control.Process.ExportStream(
		timeout, control.ProcessExportRequest{
			ProcessIDs: []api.UUID{pID},
			ExportLogs: true,
		}, func(ctx context.Context, chunks control.OctetStreamList) {
			for _, chunk := range chunks.Chunks {
				b = append(b, chunk.Data...)
			}
		}))
	var exportDataList []ProcessExportData
	require.NoError(t, yaml.Unmarshal(b, &exportDataList))

	exportData := exportDataList[0]
	assert.Equal(t, 2, len(exportData.Logs))
	assert.Equal(t, api.LogLevel_WARNING.String(), exportData.Logs[0].Message.Level.String())
	assert.Equal(t, "Actor too busy. Waiting 2s before next attempt. Reason given: hitCount 1", exportData.Logs[0].Message.Text)
	assert.Equal(t, api.LogLevel_WARNING.String(), exportData.Logs[1].Message.Level.String())
	assert.Equal(t, "Actor too busy. Waiting 4s before next attempt. Reason given: hitCount 2", exportData.Logs[1].Message.Text)
}
