// Copyright 2022 Orus.io. All rights reserved.

package control

import (
	"context"
	"runtime"

	"xbus.io/go-xbus/v3/api/control"
)

type debugServer struct {
	*Service
}

func (s debugServer) Info(
	ctx context.Context, req control.DebugInfoRequest,
	send func(control.OctetStreamList),
) (
	err error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	w := NewOctetStreamWriter(ctx, send)
	formatter := TextFormatter{w: &w}

	for _, entry := range debugInfoEntries {
		if entry.Name == req.Entry {
			if err := entry.Run(req.Args, &formatter); err != nil {
				return err
			}
		}
	}

	return nil
}

type DebugInfoEntry struct {
	Name        string
	Description string
	Help        string
	Run         func([]string, OutputFormatter) error
}

var debugInfoEntries = []DebugInfoEntry{{
	Name:        "list",
	Description: "List the available info entries",
	Help:        "",
}, {
	Name:        "stack",
	Description: "Return all the server routine stacks",
	Help:        "",
	Run: func(args []string, formatter OutputFormatter) error {
		if err := formatter.OpenSection("Stacks"); err != nil {
			return err
		}
		defer func() {
			_ = formatter.CloseSection()
		}()

		buf := make([]byte, 1024)
		for {
			n := runtime.Stack(buf, true)
			if n < len(buf) {
				buf = buf[:n]
				break
			}
			buf = make([]byte, 2*len(buf))
		}
		if err := formatter.AddValue(string(buf)); err != nil {
			return err
		}

		return nil
	},
}}

func init() {
	debugInfoEntries[0].Run = func(args []string, formatter OutputFormatter) error {
		if err := formatter.OpenSection("Available info"); err != nil {
			return err
		}
		defer func() {
			_ = formatter.CloseSection()
		}()
		if err := formatter.OpenTable("name", "description", "help"); err != nil {
			return err
		}
		defer func() {
			_ = formatter.CloseTable()
		}()

		for _, entry := range debugInfoEntries {
			if err := formatter.AddValue([]interface{}{
				entry.Name, entry.Description, entry.Help,
			}); err != nil {
				return err
			}
		}

		return nil
	}
}
