package sql

import (
	"errors"
	"fmt"
	"strings"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"github.com/slicebit/qb"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql/model"
)

func readStringList(s string) []string {
	if s == "" {
		return []string{}
	}
	return strings.Split(s, ",")
}

func writeStringList(list []string) string {
	return strings.Join(list, ",")
}

// NewPipelineStorage initialize a PipelineStorage
func NewPipelineStorage(model *model.Model, db *yago.DB, logger zerolog.Logger) *PipelineStorage {
	return &PipelineStorage{Base{
		model: model,
		db:    db,
		logger: logger.With().
			Str("storage", "pipeline").
			Str("implementation", "sql").
			Logger(),
	}}
}

// PipelineStorage implements storage.PipelineStorage
type PipelineStorage struct {
	Base
}

func unmarshalSourceMatchEventTypes(s string) []string {
	if s == "" {
		return nil
	}
	return strings.Split(s, ",")
}

func dbNodeToStorage(dbNode model.GraphNode) storage.Node {
	node := storage.Node{
		ID:            dbNode.NodeID,
		RoleBroadcast: dbNode.RoleBroadcast,
		Inputs:        readStringList(dbNode.Inputs),
		Outputs:       readStringList(dbNode.Outputs),
	}
	switch dbNode.Type {
	case "emitter":
		node.Type = storage.EmitterNode
		node.SourceMatch = &storage.SourceMatch{
			EventTypes: unmarshalSourceMatchEventTypes(dbNode.SourceMatchEventTypes),
			ByActor:    dbNode.SourceMatchByActor,
		}
	case "consumer":
		node.Type = storage.ConsumerNode
	case "worker":
		node.Type = storage.WorkerNode
	case "requestor":
		node.Type = storage.RequestorNode
		node.SourceMatch = &storage.SourceMatch{
			EventTypes: unmarshalSourceMatchEventTypes(dbNode.SourceMatchEventTypes),
			ByActor:    dbNode.SourceMatchByActor,
		}
	}
	for _, id := range strings.Split(dbNode.ActorIDs, ",") {
		if id != "" {
			node.ActorIDs = append(node.ActorIDs, api.MustUUIDFromString(id))
		}
	}
	for _, actor := range strings.Split(dbNode.Actors, ",") {
		if actor != "" {
			node.Actors = append(node.Actors, actor)
		}
	}
	if dbNode.Roles != "" {
		node.Roles = strings.Split(dbNode.Roles, ",")
	}
	return node
}

func parseEdgeList(str string) ([]storage.Edge, error) {
	strEdges := strings.Split(str, ",")
	var edges = make([]storage.Edge, 0, len(strEdges))
	for _, strEdge := range strEdges {
		if strEdge == "" {
			continue
		}
		edge, err := storage.EdgeFromString(strEdge)
		if err != nil {
			return nil, err
		}
		edges = append(edges, edge)
	}
	return edges, nil
}

func writeEdgeList(edges []storage.Edge) string {
	var strEdges = make([]string, len(edges))
	for i, edge := range edges {
		strEdges[i] = edge.String()
	}
	return strings.Join(strEdges, ",")
}

func dbPipelineToStorage(dbPipeline model.Pipeline, dbNodes []model.GraphNode) (storage.Pipeline, error) {
	var nodes = make([]storage.Node, len(dbNodes))
	for i, dbNode := range dbNodes {
		nodes[i] = dbNodeToStorage(dbNode)
	}
	edges, err := parseEdgeList(dbPipeline.Edges)
	if err != nil {
		return storage.Pipeline{}, err
	}
	return storage.Pipeline{
		PipelineInfo: api.PipelineInfo{
			ID:      api.UUID(dbPipeline.ID),
			Name:    dbPipeline.Name,
			Version: dbPipeline.Version,
			Status:  dbPipeline.Status,
		},
		Graph: storage.Graph{
			Nodes: nodes,
			Edges: edges,
		},
	}, nil
}

var storageNodeTypeToDB = []string{"emitter", "consumer", "worker", "requestor"}

func storageNodeToDB(node storage.Node) model.GraphNode {
	var strActorIDs = make([]string, len(node.ActorIDs))
	for i, id := range node.ActorIDs {
		strActorIDs[i] = id.String()
	}

	var sourceMatchEventTypes string
	if (node.Type == storage.EmitterNode ||
		node.Type == storage.RequestorNode) &&
		node.SourceMatch != nil &&
		len(node.SourceMatch.EventTypes) != 0 {
		sourceMatchEventTypes = strings.Join(node.SourceMatch.EventTypes, ",")
	}

	var sourceMatchByActor bool
	if (node.Type == storage.EmitterNode ||
		node.Type == storage.RequestorNode) &&
		node.SourceMatch != nil {
		sourceMatchByActor = node.SourceMatch.ByActor
	}

	return model.GraphNode{
		NodeID:                node.ID,
		Type:                  storageNodeTypeToDB[node.Type],
		ActorIDs:              strings.Join(strActorIDs, ","),
		Actors:                strings.Join(node.Actors, ","),
		Roles:                 writeStringList(node.Roles),
		RoleBroadcast:         node.RoleBroadcast,
		SourceMatchEventTypes: sourceMatchEventTypes,
		SourceMatchByActor:    sourceMatchByActor,
		Inputs:                writeStringList(node.Inputs),
		Outputs:               writeStringList(node.Outputs),
	}
}

func storagePipelineToDB(pipeline storage.Pipeline) (model.Pipeline, []model.GraphNode) {
	var dbNodes = make([]model.GraphNode, len(pipeline.Graph.Nodes))
	for i, node := range pipeline.Graph.Nodes {
		dbNodes[i] = storageNodeToDB(node)
	}
	return model.Pipeline{
		Base:    model.Base{ID: uuid.UUID(pipeline.ID)},
		Name:    pipeline.Name,
		Version: pipeline.Version,
		Status:  pipeline.Status,
		Edges:   writeEdgeList(pipeline.Graph.Edges),
	}, dbNodes
}

// Load loads a graph from storage
func (s *PipelineStorage) Load(id api.UUID) (storage.Pipeline, bool) {
	tx := s.MustBegin()
	defer tx.BlindRollback()
	var dbPipeline model.Pipeline
	err := tx.Query(s.model.Pipeline).Get(&dbPipeline, uuid.UUID(id))
	if errors.Is(err, yago.ErrRecordNotFound) {
		return storage.Pipeline{}, false
	}
	must(err)

	var dbNodes []model.GraphNode
	must(
		tx.Query(s.model.GraphNode).
			Where(s.model.GraphNode.PipelineID.Eq(uuid.UUID(id))).
			All(&dbNodes))

	pipeline, err := dbPipelineToStorage(dbPipeline, dbNodes)
	must(err)

	return pipeline, true
}

// Save saves a pipeline to storage
func (s *PipelineStorage) Save(pipeline storage.Pipeline) (api.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbPipeline model.Pipeline

	if !pipeline.ID.IsZero() {
		if err := tx.Query(s.model.Pipeline).
			Where(s.model.Pipeline.ID.Eq(uuid.UUID(pipeline.ID))).
			One(&dbPipeline); err != nil {
			if errors.Is(err, yago.ErrRecordNotFound) {
				err = fmt.Errorf("No such pipeline: %s", pipeline.ID)
				return api.UUID{}, err
			}
			must(err)
		}
		if dbPipeline.Status != api.PipelineInfo_DRAFT {
			return api.UUID{}, storage.ErrInvalidPipelineStatus
		}
	} else {
		if err := tx.Query(s.model.Pipeline).
			Where(
				s.model.Pipeline.Name.Eq(pipeline.Name),
				s.model.Pipeline.Version.Eq(pipeline.Version)).
			One(&dbPipeline); err != nil {
			if !errors.Is(err, yago.ErrRecordNotFound) {
				must(err)
			}
		} else {
			if dbPipeline.Status != api.PipelineInfo_DRAFT {
				return api.UUID{}, storage.ErrInvalidPipelineStatus
			}
		}
	}

	isNewPipeline := dbPipeline.ID.IsZero()

	nextPipeline, dbNodes := storagePipelineToDB(pipeline)

	dbPipeline.Name = nextPipeline.Name
	dbPipeline.Version = nextPipeline.Version
	dbPipeline.Edges = nextPipeline.Edges

	if isNewPipeline {
		must(
			tx.Insert(&dbPipeline))
	} else {
		must(
			tx.Update(
				&dbPipeline,
				model.PipelineName,
				model.PipelineVersion,
				model.PipelineEdges,
			))
	}

	var dbNodeIDs = make([]interface{}, len(dbNodes))

	q := tx.Query(s.model.GraphNode).
		Select(s.model.GraphNode.ID).
		Where(s.model.GraphNode.PipelineID.Eq(dbPipeline.ID))

	for i, dbNode := range dbNodes {
		dbNode.PipelineID = dbPipeline.ID
		var id int64
		if err := q.Filter(
			s.model.GraphNode.NodeID.Eq(dbNode.NodeID),
		).Scalar(&id); err != nil {
			if errors.Is(err, yago.ErrRecordNotFound) {
				must(
					tx.Insert(&dbNode))
			} else {
				must(err)
			}
		} else {
			dbNode.ID = id
			must(
				tx.Update(&dbNode))
		}
		dbNodeIDs[i] = dbNode.NodeID
	}

	_, err := tx.GetEngine().Exec(
		qb.Delete(*s.model.GraphNode.GetMapper().Table()).
			Where(qb.And(
				s.model.GraphNode.PipelineID.Eq(dbPipeline.ID),
				s.model.GraphNode.NodeID.NotIn(dbNodeIDs...))),
	)
	must(err)
	tx.MustCommit()

	return api.UUID(dbPipeline.ID), nil
}

// SetStatus changes a graph status
func (s *PipelineStorage) SetStatus(id api.UUID, status api.PipelineInfo_Status) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	if status != api.PipelineInfo_ACTIVE && status != api.PipelineInfo_INACTIVE {
		return storage.ErrInvalidPipelineStatus
	}

	if status == api.PipelineInfo_ACTIVE {
		// Deactivate any other version of the same graph
		var name string
		must(tx.Query(s.model.Pipeline).
			Where(s.model.Pipeline.ID.Eq(uuid.UUID(id))).
			Select(s.model.Pipeline.Name).
			Scalar(&name))

		_, err := tx.GetEngine().Exec(
			s.model.Pipeline.GetMapper().Table().Update().
				Values(map[string]interface{}{
					model.PipelineStatusColumnName: "INACTIVE",
				}).
				Where(qb.And(
					s.model.Pipeline.Name.Eq(name),
					s.model.Pipeline.Status.Eq("ACTIVE"),
				)),
		)
		must(err)
	}

	must(tx.Update(
		&model.Pipeline{Base: model.Base{ID: uuid.UUID(id)}, Status: status},
		model.PipelineStatus))

	tx.MustCommit()

	return nil
}

// LoadActivePipeline loads all the active graphs
func (s *PipelineStorage) LoadActivePipeline() []storage.Pipeline {
	tx := s.MustBegin()
	defer tx.BlindRollback()

	var dbPipelines []model.Pipeline
	must(
		tx.Query(s.model.Pipeline).
			Where(s.model.Pipeline.Status.Eq("ACTIVE")).
			All(&dbPipelines))

	var graphList = make([]storage.Pipeline, len(dbPipelines))

	for i, dbPipeline := range dbPipelines {
		var dbNodes []model.GraphNode
		must(
			tx.Query(s.model.GraphNode).
				Where(s.model.GraphNode.PipelineID.Eq(dbPipeline.ID)).
				All(&dbNodes))
		graph, err := dbPipelineToStorage(dbPipeline, dbNodes)
		must(err)
		graphList[i] = graph
	}
	return graphList
}

// Query searches for matching graphs
func (s *PipelineStorage) Query(name, version string, activeOnly bool) (graphs []storage.Pipeline) {
	tx := s.MustBegin()
	defer tx.BlindRollback()

	q := tx.Query(s.model.Pipeline)
	if name != "" {
		q = q.Filter(s.model.Pipeline.Name.Eq(name))
	}
	if version != "" {
		q = q.Filter(s.model.Pipeline.Version.Eq(version))
	}
	if activeOnly {
		q = q.Filter(s.model.Pipeline.Status.Eq(api.PipelineInfo_ACTIVE))
	}

	var dbPipelines []model.Pipeline
	must(q.All(&dbPipelines))

	for _, dbPipeline := range dbPipelines {
		var dbNodes []model.GraphNode
		must(
			tx.Query(s.model.GraphNode).
				Where(s.model.GraphNode.PipelineID.Eq(dbPipeline.ID)).
				All(&dbNodes))

		graph, err := dbPipelineToStorage(dbPipeline, dbNodes)
		must(err)

		graphs = append(graphs, graph)
	}
	return
}
